// Copyright (c) 2017, Ravi Teja Gudapati. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:math';
import 'dart:async';
import 'package:jaguar/jaguar.dart';
import 'package:jaguar_cache/jaguar_cache.dart';

final Cache defaultThrottleCache = new InMemoryCache(new Duration(seconds: -1));

/// Contains count information for a single throttle identifier/key
class ThrottleData {
  /// Count
  int count = 0;

  /// Start of the recorded time interval
  int milliseconds = 0;

  ThrottleData();

  ThrottleData.now(this.count)
      : milliseconds = new DateTime.now().millisecondsSinceEpoch;

  ThrottleData.make(this.count, this.milliseconds);

  /// Returns time
  DateTime time() => new DateTime.fromMillisecondsSinceEpoch(milliseconds);

  /// Increments the count
  ThrottleData inc(int milliseconds) =>
      new ThrottleData.make(count + 1, milliseconds);
}

/// Identifies a HTTP request given the [Context] object
///
/// For example identifier could be based on
///   1. IP address
///   2. Logged-in User
///   3. Session id
///   4. route
///   5. Etc
///   6. Or any combination of the above
typedef FutureOr<String> ThrottleIdMaker(Context ctx);

/// [ThrottleIdMaker] to throttle requests by remote IP
String throttleIdByIp(Context ctx) =>
    ctx.req.connectionInfo.remoteAddress.address;

/// Interceptor to rate-limit routes
class Throttle extends Interceptor {
  /// Maximum number of requests allowed for the given [interval]
  final int quota;

  /// Interval
  final Duration interval;

  /// Identification maker. Creates an identification for the given request
  final ThrottleIdMaker idMaker;

  final Cache store;

  /// Provides an opportunity to create [Response] when rate-limit exceeds
  final RouteHandlerFunc failResponse;

  Throttle(this.quota, this.interval,
      {this.idMaker: throttleIdByIp, this.failResponse, Cache store})
      : store = store ?? defaultThrottleCache;

  ThrottleState _state;

  Future<ThrottleState> pre(Context ctx) async {
    final String throttleId = await idMaker(ctx);
    final ThrottleData count = await _visit(throttleId, interval);

    _state = new ThrottleState.make(quota, count, interval);

    if (count == null) return _state;

    if (count.count > quota) {
      final Response resp = failResponse != null
          ? await failResponse(ctx)
          : new Response('Limit exceeded', statusCode: 429);

      throw _state.setHeaders(resp);
    }

    return _state;
  }

  Response post(Context ctx, Response incoming) => _state.setHeaders(incoming);

  Future<ThrottleData> _visit(String identifier, Duration interval) async {
    ThrottleData ret = new ThrottleData.now(1);

    ThrottleData data;

    try {
      data = await store.read(identifier);
    } catch(e) {
      if(e != cacheMiss) rethrow;
    }

    final now = new DateTime.now();

    if (data == null) {
      ret = new ThrottleData.now(1);
    } else if (now.difference(data.time()) < interval) {
      ret = data.inc(data.milliseconds);
    } else {
      ret = new ThrottleData.now(1);
    }

    await store.upsert(identifier, ret);
    return ret;
  }
}

class ThrottleState {
  final int quota;

  final int remaining;

  final int resetsIn;

  final int retryAfter;

  ThrottleState(this.quota, this.remaining, this.resetsIn, this.retryAfter);

  factory ThrottleState.make(int quota, ThrottleData data, Duration interval) {
    if (data == null) {
      return new ThrottleState(quota, quota, interval.inSeconds, -1);
    }

    int remaining = max(0, quota - data.count);
    int resetsIn = max(0, data.time().difference(new DateTime.now()).inSeconds);
    int retryAfter = -1;
    if (quota <= data.count) {
      retryAfter = max(0,
          data.time().add(interval).difference(new DateTime.now()).inSeconds);
    }
    return new ThrottleState(quota, remaining, resetsIn, retryAfter);
  }

  Response setHeaders(Response resp) {
    resp.headers.set('X-RateLimit-Limit', quota.toString());
    resp.headers.set('X-RateLimit-Remaining', remaining.toString());
    resp.headers.set('X-RateLimit-Reset', resetsIn.toString());
    if (retryAfter >= 0) resp.headers.set('Retry-After', retryAfter.toString());
    return resp;
  }
}
